# -*- encoding: utf-8 -*-

import warnings
import numpy as np
import pandas as pd
import geopandas as gpd

from copy import deepcopy
from tqdm import tqdm as TQ
from shapely.geometry import Point

from ..libs import *
from ..config import UNITS, CRS, METRICS
from ..commons.DistanceFunctions import HaversineDistance
from ..commons.GeographicalFunctions import create_circle_polygon
from ..exceptions import EncounteredNaN, DuplicateRecords, ZeroDistanceNeighbor, ValueWarning

### --- Helper Function(s) --- ###
def _type_of_comparison(data):
	'''Given the Data, it Finds either it Falls in:
		Case-I Execution  : returns 'single_file'; or
		Case-II Execution : returns 'multi_files'

	Also, this returns a deepcopy() of the Original Data, and checks the DataType!
	'''
	data = deepcopy(data)

	if type(data) == pd.DataFrame:
		return data, 'single_file'
	elif type(data) == list:
		for d in data: # Checking if both the Data is Correct or Not
			if type(d) != pd.DataFrame:
				raise TypeError(f'Expects data to be pd.DataFrame or [pd.DataFrame, pd.DataFrame], but got {type(d)}')
		return data, 'multi_files'
	elif type(data) == tuple:
		raise TypeError('tuple are immutable in Python. A Simple list of the Same, should fix the Problem.')

	raise TypeError(f'Expects data to be pd.DataFrame or [pd.DataFrame, pd.DataFrame], but got {type(data)}')

def _validate_data(_check_data):
	'''Validate Data according to the Requirement'''
	dSize = _check_data.shape[0]
	ValidationPassed = True
	_check_data = deepcopy(_check_data)

	_check_data.dropna(inplace = True)
	if _check_data.shape[0] != dSize:
		warnings.warn(f'Encounterd {dSize - _check_data.shape[0]} NaN-Values, Removed from Analysis', EncounteredNaN)
		dSize = _check_data.shape[0]

	_check_data.drop_duplicates(inplace = True)
	if _check_data.shape[0] != dSize:
		warnings.warn(f'Encounterd {dSize - _check_data.shape[0]} Duplicate-Values, Removed from Analysis', DuplicateRecords)

	return _check_data

def _generate_polygon_data(data, loc, max_dist_arr):
	'''Generate Polygon-Data with gpd.GeoDataFrame()'''
	_vals = np.array(list(data[loc].values))

	try:
		_circle_polygon = list(map(create_circle_polygon, _vals[:, 0], _vals[:, 1], max_dist_arr))
	except ValueError as err:
		raise ValueError(f'Possibly Data contains NaN\n{err}')
	except RuntimeError:
		# https://github.com/pyproj4/pyproj/issues/202
		# Err. Msg. : RuntimeError: b'tolerance condition error'
		raise RuntimeError(f'PyProj version <= 1.9.6 behaves differently than >= 2.1.0')

	data['polygons'] = _circle_polygon
	_PolygonData = deepcopy(gpd.GeoDataFrame(data, crs = CRS, geometry = data['polygons']))
	data.drop(columns = ['polygons'], inplace = True)

	return _PolygonData

def _generate_point_data(data, loc):
	'''Generate Point-Data with gpd.GeoDataFrame()'''
	data['points'] = data[loc].apply(lambda x: Point(x[0], x[1]))
	_PointData = deepcopy(gpd.GeoDataFrame(data, crs = CRS, geometry = data['points']))
	data.drop(columns = ['points'], inplace = True)

	return _PointData

### --- Main Functions --- ###
def NearestNeighbor (
		data     : pd.DataFrame or [pd.DataFrame, pd.DataFrame], # Max. of TWO DataFrame is allowed for Comparison
		num_nbr  : int = 0,
		max_dist : int = 10,
		**kwargs
	) -> pd.DataFrame:
	'''NearestNeighbor can locate (N) number of Neighbors for a given Point/Location within a Search-Area
	
	Types of Comparison
		1. Within Same-File/DataFrame : Given x-Points/Locations in a DataFrame, it searches the Nearest One, excluding SELF
		2. Within Two Different Files : Foreach Point in File-A, it searches for the Nearest One from File-B

	Thus, there can be two type(s) of Data Input: either only a single DataFrame or a List of Two DataFrames.

	Parameters
	----------
		: param data : List of DataFrame(s) for either Case-I Execution Type, or Case-II Execution Type
		               By Default, if only a Single DataFrame is Passed then Case-I is Executed, else Case-II
		               (which are auto-assigned). Default Column Names for Each of the DataFrame is as Below:
		                   >> UIN : Unique Identification No. for the Point/Location
		                   >> loc : Location of the Point either
		                            (i) Point in n-Dimension [x1, x2, x3, ... xn], where n = No. of Diemnsion; or
		                            (ii) Actual Point on Earth in the Format [longitude, latitude]
		                            Location Object can be either a list or a tuple
		: param num_nbr : Number of Neighbor to Find. By Default num_nbr = 0, representing all the Neighbors
		                  num_nbr can be (i) unitless - for Point in n-Dimensions, or (ii) with unit for others
		                  Set the Data Unit using the kwargs as Defined below
		: param max_dist : Max. Distance within which a Neighbor has to be Located. Currently it does not understand infinity!

	Keyword Arguments
	-----------------
		: param point_type  : (str) Type of the Point, either
		                      (i) 'n-dim' - Stating a n-dimensional Point
		                      (ii) 'geo-loc' - An Actual Location on the Earth
		                      Default: 'geo-loc'
		: param input_unit  : (str) Can be ['m', 'km', 'rad', 'none']. Default 'km'
		: param output_unit : (str) Can be ['m', 'km', 'rad', 'none']. Default 'km'
		: param dist_metric : (str) Distance Metric to Use ['euclidean', 'manhattan', 'haversine']
		                      If point_type == geo-loc, then by Default haversine is used, else euclidean

		Apart from this, the Column Names can be assigned on-th-go with the following Arguments:
		: param UIN : Default 'UIN'
		: param loc : Default 'loc'

		: param keep_blanks : (bool) Keep UIN which do not have any Neighbor within the Search Radius. Default True
		                      By Default, for Blank-UINs, the NBR-UIN (with other corresponding values) is set to NaN
		                      If required, this can be changed by passing a list() as given in fillna-command
		: param fillna      : (list) fillna values with required value. Format is [bool, '<value>']
		                      By Default fillna [False, np.nan]

	NOTE: The Entire workflow is with datum = WGS84 i.e. CRS = EPSG:4326 - any other System will/might Create a Problem.
	NOTE: If the Data contains any illegal chars, it is advised to remove from Analysis, as all the Values are Not Validated.
	RETURNS : pd.DataFrame of [UIN, loc, f'NBR_{UIN}', f'NBR_{loc}', 'rnk'] # where rnk = Rank is not present for num_nbr = 1
	'''
	data, _exec_type = _type_of_comparison(data)

	### --- Getting the Keyword Arguments, and Validate --- ###
	input_unit  = kwargs.get('input_unit', 'km')
	output_unit = kwargs.get('output_unit', 'km')

	if input_unit not in UNITS: raise ValueError(f'input_unit = {input_unit} is not Understood.')
	if output_unit not in UNITS: raise ValueError(f'output_unit = {output_unit} is not Understood.')

	UIN = kwargs.get('UIN', 'UIN') # the value can be anything
	loc = kwargs.get('loc', 'loc') # the value can be anything

	point_type  = kwargs.get('point_type', 'geo-loc')

	if point_type == 'geo-loc':
		dist_metric = kwargs.get('dist_metric', 'haversine')
	elif point_type == 'n-dim':
		dist_metric = kwargs.get('dist_metric', 'euclidean')
	else:
		raise ValueError(f'{point_type} is not Understood.')

	if dist_metric not in METRICS: raise ValueError(f'dist_metric = {dist_metric} is not Understood.')

	keep_blanks = kwargs.get('keep_blanks', True)
	fillna      = kwargs.get('fillna', [False, np.nan])

	### --- Minimal Validation of Data --- ###
	timer = profilingFunction(funcName = 'Generating shapely Objects')
	if _exec_type == 'single_file':
		data = _validate_data(data) # Validating Records
		max_dist_arr = [max_dist * 1000 for _ in range(data.shape[0])]

		### --- create_circle_polygon and create_points --- ###
		PointData = _generate_point_data(data, loc)
		PolygonData = _generate_polygon_data(data, loc, max_dist_arr)

	elif _exec_type == 'multi_files':
		FileA = _validate_data(data[0])
		FileB = _validate_data(data[1])
		max_dist_arr = [max_dist * 1000 for _ in range(FileA.shape[0])]

		### --- create_circle_polygon and create_points --- ###
		PointData = _generate_point_data(FileB, loc)
		PolygonData = _generate_polygon_data(FileA, loc, max_dist_arr)

	timer.checkPoint()

	### --- Generating Neighbors List --- ###
	timer = profilingFunction(funcName = 'Joining Data with gpd.sjoin()')
	all_neighbor = gpd.sjoin(PolygonData, PointData, how = 'left', op = 'contains').reset_index()
	del PolygonData, PointData # House-Keeping!

	all_neighbor.drop(columns = ['index', 'index_right', 'geometry', 'points', 'polygons'], inplace = True)
	all_neighbor.rename(columns = {
			f'{UIN}_left'  : UIN,
			f'{loc}_left'  : loc,
			f'{UIN}_right' : f'NBR_{UIN}',
			f'{loc}_right' : f'NBR_{loc}'
		}, inplace = True)

	all_neighbor = all_neighbor[all_neighbor[UIN] != all_neighbor[f'NBR_{UIN}']] # Removing Same Site(s)
	timer.checkPoint()

	### --- keep_blanks Execution --- ###
	if keep_blanks:
		print(f'NBR_{UIN} == NaN - denotes a Location w/o a Neighbor')
		if _exec_type == 'single_file':
			_blank_UIN = np.setdiff1d(data[UIN].unique(), all_neighbor[UIN].unique())
			_blank_UIN = data[data[UIN].isin(_blank_UIN)][[UIN, loc]].reset_index()

			del data # data is no Longer Required, clearing Memory!
		elif _exec_type == 'multi_files':
			_blank_UIN = np.setdiff1d(FileA[UIN].unique(), all_neighbor[UIN].unique())
			_blank_UIN = FileA[FileA[UIN].isin(_blank_UIN)][[UIN, loc]].reset_index()

			del data, FileA, FileB # data is no Longer Required, clearing Memory!

		_blank_UIN.drop(columns = ['index'], inplace = True)
		_blank_UIN['rnk'] = [0] * _blank_UIN.shape[0] # this is essential for Filtering Purposes

	startPoints  = all_neighbor[loc].values
	targetPoints = all_neighbor[f'NBR_{loc}'].values

	_calculated_distance = []
	for s, t in TQ(zip(startPoints, targetPoints), desc = f'Calculating Distance with {dist_metric} Metric'):
		_calculated_distance.append(HaversineDistance(s, t, output_unit = output_unit))

	all_neighbor[f'Calculated Distance ({output_unit})'] = _calculated_distance
	if all_neighbor[f'Calculated Distance ({output_unit})'].min() == 0:
		_zero_dist_nbr = all_neighbor[all_neighbor[f'Calculated Distance ({output_unit})'] == 0].shape[0]
		warnings.warn(f'Found {_zero_dist_nbr}/{all_neighbor[UIN].nunique()} Neighbors at 0 units Distance')

	timer = profilingFunction(funcName = 'Calculating Ranks')
	all_neighbor['rnk'] = all_neighbor.groupby([UIN])[f'Calculated Distance ({output_unit})'].rank(method = 'dense')
	if keep_blanks:
		all_neighbor = pd.concat([all_neighbor, _blank_UIN], sort = True, ignore_index = True)

	all_neighbor.sort_values(by = [UIN, 'rnk'], inplace = True)

	if fillna[0]:
		try:
			all_neighbor.fillna(fillna[1], inplace = True)
		except ValueError as err:
			warnings.warn(f'{fillna[1]} is not accepted for pd.fillna(): {err}. Converting {fillna[1]} to str({fillna[1]})')
			all_neighbor.fillna(str(fillna[1]), inplace = True)

	all_neighbor = all_neighbor[[UIN, loc, f'NBR_{UIN}', f'NBR_{loc}', f'Calculated Distance ({output_unit})', 'rnk']]
	timer.checkPoint()

	if num_nbr == 0:
		return all_neighbor

	return all_neighbor[all_neighbor.rnk <= num_nbr]


# -*- encoding: utf-8 -*-

import warnings
from platform import system, release

from ..exceptions import LimitedFunctionality

def check_imports(module_names):
	_unavailable = []
	for i in module_names:
		try:
			__import__(i)
		except ImportError:
			_unavailable.append(i)

	return _unavailable

class _env_setup:
	"""Sets the OS-Options"""
	def __init__(self):
		self.OSName    = system()
		self.OSVersion = release()

		self.point_func = ['pandas', 'fiona', 'shapely', 'pyproj', 'geopandas']
		self.line_of_st = self.point_func + ['rasterio']

		self.available_libs = ['numpy']

	@property
	def _point_func(self):
		return self.point_func

	@property
	def _line_of_st(self):
		return self.line_of_st

	@property
	def _available_libs(self):
		return self.available_libs

	@property
	def is_threading_possible(self):
		if 'linux' in self.OSName.lower():
			return True
		elif 'darwin' in self.OSName.lower():
			return True
		else:
			return False # Multi-Threading Creates problem in Windows-Environment

	@property
	def is_tensorflow_available(self):
		try:
			__import__('tensorflow')
			self.available_libs.append('tensorflow')
			return True
		except ImportError:
			return False

	@property
	def is_sklearn_available(self):
		try:
			__import__('sklearn')
			self.available_libs.append('scikit-learn')
			return True
		except ImportError:
			return False

	def module_functionality(self):
		__global_options__ = ['point-function', 'line-of-sight']
		# Check what are the available pkgs - on which global-functionality is set
		AVLBL_OPTIONS = ['commons']
		for opts, libs in zip(__global_options__, [self.point_func, self.line_of_st]):
			_check = check_imports(libs)
			if not _check:
				AVLBL_OPTIONS.append(opts)
				self.available_libs += libs
			else:
				warnings.warn(f"{opts} Functionality is NOT Available, as {_check} are Required.", LimitedFunctionality)
				_input = input('Do you want to Continue? (Y/n) ')
				if (_input == 'n') or (_input == 'N'):
					raise ImportError(f'{_check} required for {opts}')

		return AVLBL_OPTIONS

OSOptions = _env_setup() # Create a Object - that can be accessed from Everywhere!


# -*- encoding: utf-8 -*-

__author__       = 'Debmalya Pramanik'
__author_email__ = ""

__version__   = "0.0.1"
__copyright__ = "Copyright (c) 2020 Debmalya Pramanik"

__status__    = "dev"
__docformat__ = "camelCasing"

# Let's Check for the Dependencies
hardDependencies    = ['os', 'sys', 'math', 'random', 'numpy']
missingDependencies = []

for dependency in hardDependencies:
	try:
		__import__(dependency)
	except ImportError:
		missingDependencies.append(dependency)

if missingDependencies:
	raise ImportError('Required Dependencies {}'.format(missingDependencies))

print("Setting up agine-Environment...")
from .api import *


# -*- encoding: utf-8 -*-

import pyproj
import warnings
import numpy as np
from math import degrees
from shapely import geometry
from functools import partial
from shapely.ops import transform
from shapely.geometry import Point, Polygon

from ..config import R, DATUM
from ..exceptions import TypeWarning

# Check https://gis.stackexchange.com/questions/121256/ and
# Check https://gis.stackexchange.com/questions/289044/ for More Information
# NOTE : this converts everything into 'm'

def create_circle_polygon(centerLon : int or float, centerLat : int or float, radius : int or float = 10000) -> Polygon:
	'''Create a Circle from the given [longitude, latitude] with the given Radius

	: param centerLat : Center Latitude - in WGS84
	: param centerLon : Center Longitude - in WGS84

	: param radius    : Radius of the Circle, in 'm' ONLY. Default 10000 = 10 km

	Returns shapely.Polygon Object
	'''
	local_azimuthal_projection = f"+proj=aeqd +R={R} +units=m +lat_0={centerLat} +lon_0={centerLon}"

	WGS84_to_AEQD = partial( # AEQD = Azimuthal Equidistance Projection
			pyproj.transform,
			pyproj.Proj(local_azimuthal_projection),
			pyproj.Proj(f"+proj=longlat +datum={DATUM} +no_defs")
		)

	AEQD_to_WGS84 = partial(
			pyproj.transform,
			pyproj.Proj(local_azimuthal_projection),
			pyproj.Proj(f"+proj=longlat +datum={DATUM} +no_defs")
		)

	point = Point(float(centerLon), float(centerLat)) # Center-Point
	point = transform(WGS84_to_AEQD, point)

	circular_poly = point.buffer(radius)
	circular_poly = transform(AEQD_to_WGS84, circular_poly)

	return circular_poly

def calculate_bearing_angle(startPoint : tuple, targetPoint : tuple, **kwargs) -> float:
	'''Calculates the Bearing Angle b/w two Points

	Mathematically, the Bearing Angle b/w two-points is Calculated by:
		θ = atan2(sin(Δlong).cos(lat2), cos(lat1).sin(lat2) − sin(lat1).cos(lat2).cos(Δlong))

	: param startPoint  : Starting Point in the format (longitude, latitude)
	: param targetPoint : Target/End Point in the format (longitude, latitude)

	Keyword Arguments
		: param input_unit  : (str) Output Unit which can be ['deg', 'rad']. Default deg
		: param output_unit : (str) Output Unit which can be ['deg', 'rad']. Default deg

	Return bearing_angle (float) in desired Units.
	'''
	input_unit  = kwargs.get('input_unit', 'deg')
	output_unit = kwargs.get('output_unit', 'deg')

	if input_unit not in ['deg', 'rad']: raise ValueError(f'{input_unit} is not an accepted input_unit.')
	if output_unit not in ['deg', 'rad']: raise ValueError(f'{output_unit} is not an accepted output_unit.')

	if (type(startPoint) != tuple) or (type(targetPoint) != tuple):
		startPoint  = tuple(startPoint)
		targetPoint = tuple(targetPoint)

		warnings.warn(f'Expects tuple() got {type(targetPoint)} and {type(targetPoint)}. Working with list() is Not-Advised', TypeWarning)

	if input_unit == 'deg':
		lon1, lat1, lon2, lat2 = map(np.radians, [startPoint[0], startPoint[1], targetPoint[0], targetPoint[1]])
	else:
		lon1, lat1, lon2, lat2 = startPoint[0], startPoint[1], targetPoint[0], targetPoint[1]

	dLon = lon2 - lon1
	theta = np.arctan2(np.sin(dLon) * np.cos(lat2), np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(dLon))

	if output_unit == 'rad':
		return theta

	theta = degrees(theta)
	theta = (theta + 360) % 360

	return theta # this is in degrees

def calculate_point_at_bearing(point : tuple, distance : int or float, theta : int or float, **kwargs) -> list:
	'''Calculate the Geographich Point (lon, lat) given an Initial Point, Bearing Angle and Distance

	: param point    : Satrting/take-off Point (lon, lat)
	: param theta    : Bearing Angle at which the Point is Required
	: param distance : Required Distance b/w the startPoint and the Calculated Point. Distance Unit is in meters
	                   Change the input unit using the keyword input_distance_unit

	Keyword Arguments
		: param input_unit  : (str) Output Unit which can be ['deg', 'rad']. Default deg
		: param output_unit : (str) Output Unit which can be ['deg', 'rad']. Default deg

		: param input_distance_unit : (str) Input Distance Unit, which can be ['m', 'km']. Default 'm'
		                              **NOTE the similar argument name: input_unit and input_distance_unit

	Returns (lon, lat) in Degress
	'''
	from ..config import R # Unable to access the value R during import?

	input_unit  = kwargs.get('input_unit', 'deg')
	output_unit = kwargs.get('output_unit', 'deg')

	if input_unit not in ['deg', 'rad']: raise ValueError(f'{input_unit} is not an accepted input_unit.')
	if output_unit not in ['deg', 'rad']: raise ValueError(f'{output_unit} is not an accepted output_unit.')

	input_distance_unit = kwargs.get('input_distance_unit', 'm')
	if input_distance_unit not in ['m', 'km']: raise ValueError(f'{input_distance_unit} is not an accepted input_distance_unit.')

	if input_distance_unit == 'km':
		R /= 1000 # Change R to kilometer

	if type(point) != tuple:
		point  = tuple(point)
		warnings.warn(f'Expects tuple() got {type(point)}. Working with list() is Not-Advised', TypeWarning)

	if input_unit == 'deg':
		theta    = np.deg2rad(theta)
		lon, lat = map(np.radians, point)
	else:
		lon, lat = point

	lat2 = np.arcsin(np.sin(lat) * np.cos(distance / R) + np.cos(lat) * np.sin(distance / R) * np.cos(theta))
	lon2 = lon + np.arctan2(np.sin(theta) * np.sin(distance / R) * np.cos(lat), np.cos(distance / R) - np.sin(lat) * np.sin(lat))

	if output_unit == 'rad':
		return lon2, lat2

	lon2, lat2 = map(np.degrees, [lon2, lat2])
	return lon2, lat2 # this is in Degrees


# -*- encoding: utf-8 -*-

import warnings

class LimitedFunctionality(Warning):
	"""Warning is Raised when Limited-Working Functionality is Available"""

class TypeWarning(Warning):
	"""Warning is Raised if any of the argument-type is Wrong, but can be Fixed"""

class EncounteredNaN(Warning):
	"""Warning if data contains NaN Values"""

class DuplicateRecords(Warning):
	"""Warning is raised if DataFrame contains Duplicate Records"""

class ZeroDistanceNeighbor(Warning):
	"""Warning is Raised if the Neighbor is at 0 units Distance from Parent"""

class ValueWarning(Warning):
	"""A Raised ValueError which is taken-cared by str()/other Parsing Methods"""

class VersionWarning(Warning):
	"""Warning is Raised when a PKG is available, but does not meet the Min. PKG Version Requirement"""


# -*- encoding: utf-8 -*-

import warnings
import numpy as np
from math import radians

from ..config import R
from ..api import OSOptions
from ..exceptions import TypeWarning

if OSOptions.is_sklearn_available:
	from sklearn.metrics.pairwise import haversine_distances # Using SKLearn for Calculating Haversine

def EuclideanDistance(startPoint : np.ndarray, targetPoint : np.ndarray) -> float:
	'''Calculates Eculidean Distance b/w Two Points in n-Dimensional Space

	: param startPoint  : Starting Point (P)
	: param targetPoint : Target Point, or Point-Q
	'''
	if (type(startPoint) != np.ndarray) or (type(targetPoint) != np.ndarray):
		warnings.warn(f'Expects np.ndarray got {type(startPoint)} and {type(targetPoint)}, Attempting a Fix.', TypeWarning)
		startPoint  = np.array(startPoint)
		targetPoint = np.array(targetPoint)

	return np.sqrt(np.sum((startPoint - targetPoint) ** 2))

def ManhattanDistance(startPoint : np.ndarray, targetPoint : np.ndarray) -> float:
	'''Calculates Manhattan Distance b/w Two Points in n-Dimensional Space

	: param startPoint  : Starting Point (P)
	: param targetPoint : Target Point, or Point-Q
	'''
	if (type(startPoint) != np.ndarray) or (type(targetPoint) != np.ndarray):
		warnings.warn(f'Expects np.ndarray got {type(startPoint)} and {type(targetPoint)}, Attempting a Fix.', TypeWarning)
		startPoint  = np.array(startPoint)
		targetPoint = np.array(targetPoint)

	return sum([abs(i) for i in (startPoint - targetPoint)])

def HaversineDistance(startPoint : list, targetPoint : list, **kwargs) -> float:
	'''Calculate Haversine Distance

	: param startPoint  : Starting Point (P). Format [longitude, latitude]
	: param targetPoint : Target Point, or Point-Q. Format [longitude, latitude]

	Keyword Arguments
	-----------------
		: param input_unit  : (str) Unit of the Latitude and Longitude (in deg or rad). Default 'deg'
		: param output_unit : (str) Output unit (in rad, m, km). Default 'km'
	'''
	input_unit  = kwargs.get('input_unit', 'deg')
	output_unit = kwargs.get('output_unit', 'km')

	# Check argument(s)
	if input_unit not in ['deg', 'rad']:
		raise ValueError(f'{input_unit} is invalid.')

	if output_unit not in ['m', 'km', 'rad']:
		raise ValueError(f'{output_unit} is invalid.')

	if input_unit == 'deg':
		startPoint_rad  = [radians(i) for i in startPoint]
		targetPoint_rad = [radians(i) for i in targetPoint]

	if OSOptions.is_sklearn_available:
		dist = haversine_distances([startPoint_rad, targetPoint_rad])
		dist = dist[0][1] # since this returns a np.ndarray or 2x2
	else:
		dist = _haversine_wo_sklearn_(*startPoint, *targetPoint)

	if output_unit == 'rad':
		return dist
	elif output_unit == 'm':
		return dist * R

	return dist * R / 1000

def _haversine_wo_sklearn_(lon1 : int or float, lat1 : int or float, lon2 : int or float, lat2 : int or float):
	'''Calculate Haversine, if scikit-learn Package is unavailable'''
	dlat = lat2 - lat1 # delta Latitude
	dlon = lon2 - lon1 # delta Longitude

	a = np.sin(dlat/2.0)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2.0)**2
	c = 2 * np.arcsin(np.sqrt(a)) # this is in Radian, no need to Change

	return c


# -*- encoding: utf-8 -*-

"""API is useful for (1) Exposing Functionalities, with init-time Option Registrations, and (2) Setting up agine-Environment"""

# (2) agine-Environment Setup
from .OSOptions import OSOptions
AVLBL_OPTIONS = OSOptions.module_functionality()
print(f'\tDetected OS            : {OSOptions.OSName}-{OSOptions.OSVersion}')
print(f'\tscikit-learn Options   : {OSOptions.is_sklearn_available}')
print(f'\tThreading Availibilty  : {OSOptions.is_threading_possible}')
print(f'\tTensorflow Availibilty : {OSOptions.is_tensorflow_available}')

### --- Check Lib-Version Requirement --- ###
import warnings
from ..config import _min_ver_required_
from ..exceptions import VersionWarning, LimitedFunctionality

try:
	import pkg_resources
	_get_available_libs = OSOptions._available_libs
	print(f'Available PKGS:      {_get_available_libs}')
	print(f'Available Functions: {AVLBL_OPTIONS}')

	for pkgName in _get_available_libs:
		_pkgVer_ = pkg_resources.get_distribution(pkgName).version
		_pkgVerF = float('.'.join(_pkgVer_.split('.')[:2]))
		_minReq_ = _min_ver_required_(pkgName)

		try:
			if _minReq_[0] > _pkgVerF:
				warnings.warn(f'{pkgName} Requires {_minReq_[1]} or Above, Got {_pkgVer_}', VersionWarning)
		except TypeError as err: # Required Improvevement, Issue #3
			warnings.warn(f'Unable to Check {pkgName}, got TypeError {err}.\nGot Version: ?{_pkgVer_}, Req.: ?{_minReq_[1]}',
				LimitedFunctionality)

	del pkg_resources, _get_available_libs, _pkgVer_, _pkgVerF, _minReq_
except ImportError:
	warnings.warn('Unable to Check pkg_versions as pkg_resources is unavailable', LimitedFunctionality)

del warnings, _min_ver_required_, VersionWarning, LimitedFunctionality

# (1) init-Time Option Registrations
from ..commons import * # this is alaways Available!

if 'point-function' in AVLBL_OPTIONS:
	from ..core.point_function import *
	from ..commons.GeographicalFunctions import * # though this is under commons, but requires Shapely and pyProj for Operations

if 'line-of-sight' in AVLBL_OPTIONS:
	pass


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from setuptools import setup
from setuptools import find_packages

VERSION = "0.0.2"

with open("README.md", "r") as fh:
	long_description = fh.read()

setup(
		name         = "agine",
		version      = VERSION,
		author       = "Debmalya Pramanik",
		author_email = "",

		description                   = "Functionalities for Points in n-Dimensional Space",
		long_description              = long_description,
		long_description_content_type = "text/markdown",

		url         = "https://github.com/ZenithClown/agine",
		packages    = find_packages(),
		classifiers = [
			"Programming Language :: Python :: 3",
			"License :: OSI Approved :: Apache Software License"
		],
		python_requires  = ">=3.6",
		install_requires = ["numpy"]
	)


# -*- encoding: utf-8 -*-

'''Config File - for CONSTANTS'''
R = 6371000 # Radius of the Earth

CRS     = {'init' : 'epsg:4326'}
DATUM   = 'WGS84'
UNITS   = ['m', 'km', 'rad', 'none']
METRICS = ['euclidean', 'manhattan', 'haversine']

### --- Min. Version Requirement --- ###
def _min_ver_required_(lib_name : str) -> list:
	'''Given a Python-Library Name (required by this Module) - Returns the Min. Required Version
	This is useful, to avoid any unwanted errors, for instance:
		1. geopandas <= 0.5.0 has no Atrribute gpd.points_from_xy
		2. pyproj 1.9.6 and 2.1.0 behaves differently. Issue #202 (pyProj)
		   >> https://github.com/pyproj4/pyproj/issues/202
	To mitigate this issues, this version is invoked at init-time Option Registrations - Warning users of the Same
	TODO : Changes in the setup file, to update packages. However, this might create an Issue in Windows OS
	The Format of the Version is given as lib_nam : [version, expanded_version]
	'''
	return {
		'numpy'      : lambda : [1.18, '1.18.1'],
		# Special Libraries
		'sklearn'    : lambda : [0.21, '0.21.3'],
		'tensorflow' : lambda : [2.2, '2.2.0'], # TensorFlow 1.x has some missing attributes
		# point_func Libraries
		'fiona'      : lambda : [1.8, '1.8.6'],
		'pandas'     : lambda : [0.24, '0.24.2'],
		'pyproj'     : lambda : [2.4, '2.4.2.post1'],
		'shapely'    : lambda : [1.6, '1.6.4.post2'],
		'geopandas'  : lambda : [0.5, '0.5.0'],
		# line-of-sight Libraries
		'rasterio'   : lambda : [1.1, '1.1.2']
	}.get(lib_name, lambda : f'{lib_name} is not currently required for agine')()


# -*- coding: utf-8 -*-
# Indentation: Notepad++

'''Profiling Functions: Input Parameters: startTime, send the Start Time'''

__version__ = 1.6 # Modified to Print Elapsed Time in Hour, Minute and Seconds
__author__ = "Debmalya Pramanik"
__author_email__ = "Debmalya.Pramanik@ril.com"

import time
import math

def TimeConvert(elapsedSeconds):
	convertedTime = dict()
	for multiplier, dictKey in zip([3600, 60], ['HOURs', 'MINUTEs']):
		_val = math.floor(elapsedSeconds / multiplier)
		convertedTime[dictKey] = _val
		elapsedSeconds -= _val * multiplier
		
	convertedTime['SECONDs'] = round(elapsedSeconds, 2)
	
	return f"{convertedTime['HOURs']} Hour, {convertedTime['MINUTEs']} Minutes and {convertedTime['SECONDs']} Secs."

class profilingFunction:
	def __init__(self, funcName = '', internalFunc = False):
		self.funcName = funcName
		self.internalFunc = internalFunc
		if not self.internalFunc:
			print('Start >> {funcName}, at {cTime}'.format(funcName = self.funcName, cTime = time.ctime()))
		else:
			print('\t Start >> {funcName}, at {cTime}'.format(funcName = self.funcName, cTime = time.ctime()))
		self.startTime = time.time()
		
	@property
	def elapsedTime(self):
		endTime = time.time() - self.startTime
		if endTime >= 300:
			endTime = TimeConvert(endTime)
		else:
			endTime = f'{round(endTime, 2)} Secs.'
		return endTime
			
	def checkPoint(self):
		if not self.internalFunc:
			print('\t Completed in {endTime}'.format(endTime = self.elapsedTime))
		else:
			print('\t\t  Completed in {endTime} Seconds'.format(endTime = self.elapsedTime))
			
	def __enter__(self):
		return self
		
	def __exit__(self, type, value, traceback):
		self.checkpoint()
		pass